import datetime

from graphene.types.scalars import Scalar
from graphql.language import ast


class Timestamp(Scalar):

    @classmethod
    def serialize(cls, value):
        return value.timestamp()

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.FloatValue):
            return cls.parse_value(node.value)

    @classmethod
    def parse_value(cls, value):
        return datetime.datetime.fromtimestamp(value)



import re

dashed_to_camel_regex = re.compile(r'(?!^)_([a-zA-Z])')


def dashed_to_camel(dashed_data):
    data = {}
    for key, value in dashed_data.items():
        if isinstance(value, dict):
            value = dashed_to_camel(value)

        dashed_key = dashed_to_camel_regex.sub(
            lambda match: match.group(1).upper(), key)

        data[dashed_key] = value
    return data



from graphene.types.generic import GenericScalar

from ..utils import dashed_to_camel


class CamelJSON(GenericScalar):

    @classmethod
    def serialize(cls, value):
        return dashed_to_camel(value)



import traceback
from collections import OrderedDict

from django.conf import settings
from django.contrib.auth.mixins import LoginRequiredMixin
from django.utils.translation import ugettext as _

from graphene_django.views import GraphQLView as BaseGraphQLView
from graphql.error import format_error as format_graphql_error
from graphql.error import GraphQLError
from graphql.error.located_error import GraphQLLocatedError

from . import exceptions


class GraphQLView(BaseGraphQLView):

    @staticmethod
    def format_error(error):
        if isinstance(error, GraphQLLocatedError):
            error = error.original_error

        formatted_error = format_graphql_error(error)
        graphql_error = OrderedDict([('type', error.__class__.__name__)])

        if settings.DEBUG or isinstance(error, (
            exceptions.GraphQLError,
            GraphQLError,
        )):
            graphql_error['message'] = formatted_error['message']
        else:
            # Internal errors must be masked
            graphql_error['message'] = _('Internal server error')

        if isinstance(error, exceptions.GraphQLError):
            graphql_error['code'] = error.code

            if error.error_data:
                graphql_error['data'] = error.error_data
        else:
            graphql_error['code'] = 'error'

        if 'location' in formatted_error:
            graphql_error['location'] = formatted_error['location']

        if error.__traceback__ is not None:
            info = error.__traceback__.tb_frame.f_locals.get('info')

            if info is not None:
                graphql_error['path'] = [info.field_name]
                graphql_error['operation'] = info.operation.operation

        if settings.DEBUG:
            graphql_error['trace'] = traceback.format_list(
                traceback.extract_tb(error.__traceback__))

        return graphql_error


class LoginRequiredGraphQLView(LoginRequiredMixin, GraphQLView):
    """Adding login required"""



from django.core.exceptions import ValidationError
from django.core.validators import validate_email

from graphene.types.scalars import Scalar
from graphql.language import ast

from .. import exceptions


class Email(Scalar):

    @classmethod
    def serialize(cls, value):
        return str(value)

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)

    @classmethod
    def parse_value(cls, value):
        try:
            validate_email(value)
        except ValidationError as err:
            raise exceptions.ValidationError(str(err))
        return value



from django.contrib.auth import get_user
from django.core.handlers.wsgi import WSGIRequest
from django.test import Client, RequestFactory, testcases

from graphene_django.settings import graphene_settings

from .views import GraphQLView


class GraphQLRequestFactory(RequestFactory):

    def execute(self, context, query, variables, extra):
        response = self._schema.execute(
            query,
            context_value=context,
            variable_values=variables)

        if response.errors is not None:
            response.errors = [
                GraphQLView.format_error(error)
                for error in response.errors
            ]

        return response


class GraphQLClient(GraphQLRequestFactory, Client):

    def __init__(self, **defaults):
        super().__init__(**defaults)
        self._credentials = {}
        self._schema = graphene_settings.SCHEMA

    def request(self, **request):
        request = WSGIRequest(self._base_environ(**request))

        if self.session:
            request.session = self.session
            request.user = get_user(request)
        return request

    def credentials(self, **kwargs):
        self._credentials = kwargs

    def execute(self, query, variables=None, **extra):
        extra.update(self._credentials)
        context = self.post('/', **extra)
        return super().execute(context, query, variables, extra)

    def logout(self):
        super().logout()
        self._credentials.pop('HTTP_AUTHORIZATION', None)


class GraphQLTestCase(testcases.TestCase):
    client_class = GraphQLClient



from django.utils.translation import ugettext as _

from graphene.types.scalars import Scalar
from graphql.language import ast
from graphql_relay import from_global_id

from .. import exceptions


class GlobalID(Scalar):

    @classmethod
    def serialize(cls, value):
        return str(value)

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)

    @classmethod
    def parse_value(cls, value):
        try:
            _type, _id = from_global_id(value)
        except (TypeError, ValueError):
            raise exceptions.ValidationError(_('ID cannot be resolved'))
        return int(_id)



from graphene.types.mutation import MutationOptions

from .shortcuts import get_object_or_not_found


class RetrieveOptions(MutationOptions):
    lookup_field = 'pk'
    lookup_argument = 'id'


class RetrieveMixin(object):

    @classmethod
    def __init_subclass_with_meta__(cls, _meta=None, lookup_field=None,
                                    lookup_argument=None, **options):
        if _meta is None:
            _meta = RetrieveOptions(cls)

        if lookup_field is not None:
            _meta.lookup_field = lookup_field

        if lookup_argument is not None:
            _meta.lookup_argument = lookup_argument

        super().__init_subclass_with_meta__(_meta=_meta, **options)

    @classmethod
    def get_queryset(cls, context):
        assert cls.queryset is not None, (
            '`{}` should either include a `queryset` attribute, '
            'or override the `get_queryset()` method.'.format(cls.__name__)
        )

        return cls.queryset.all()

    @classmethod
    def get_object(cls, context, **kwargs):
        assert cls._meta.lookup_argument in kwargs, (
            'Expected mutation {0} to be called with argument '
            'named "{1}". Fix your argument list, or set the '
            '`.lookup_field` on the {0}.Meta class.'
            .format(cls.__name__, cls._meta.lookup_argument)
        )

        filter_kwargs = {
            cls._meta.lookup_field: kwargs[cls._meta.lookup_argument],
        }

        queryset = cls.get_queryset(context)
        return get_object_or_not_found(queryset, **filter_kwargs)


class UpdateMixin(RetrieveMixin):

    @classmethod
    def validate(cls, context, instance, **data):
        return data

    @classmethod
    def update(cls, context, **kwargs):
        instance = cls.get_object(context, **kwargs)
        validated_data = cls.validate(context, instance, **kwargs)

        for attr, value in validated_data.items():
            setattr(instance, attr, value)

        cls.perform_update(instance)
        return instance

    @classmethod
    def perform_update(cls, instance):
        instance.save()



from django.utils.translation import ugettext as _

from graphene.types.scalars import Scalar
from graphql.language import ast
from graphql_extensions import exceptions


class Choices(Scalar):

    @classmethod
    def serialize(cls, value):
        return str(value)

    @classmethod
    def parse_literal(cls, node):
        if isinstance(node, ast.StringValue):
            return cls.parse_value(node.value)

    @classmethod
    def parse_value(cls, value):
        choices = cls.CHOICES

        if hasattr(choices, '_identifier_map'):
            choices = choices._identifier_map

        if value not in choices:
            raise exceptions.ValidationError(
                _('Invalid choice \'{}\'').format(value),
            )
        return value



from functools import wraps

from .. import exceptions

__all__ = [
    'login_required',
    'staff_member_required',
]


def context(f):
    def _context(func):
        def wrapper(*args, **kwargs):
            info = args[f.__code__.co_varnames.index('info')]
            return func(info.context, *args, **kwargs)
        return wrapper
    return _context


def login_required(f):
    @wraps(f)
    @context(f)
    def wrapper(context, *args, **kwargs):
        if context.user.is_anonymous:
            raise exceptions.NotAuthenticated()
        return f(*args, **kwargs)
    return wrapper


def staff_member_required(f):
    @wraps(f)
    @context(f)
    def wrapper(context, *args, **kwargs):
        user = context.user
        if user.is_active and user.is_staff:
            return f(*args, **kwargs)
        raise exceptions.NotAuthenticated()
    return wrapper



from django.utils.translation import ugettext_lazy as _


class GraphQLError(Exception):
    default_message = _('A server error occurred')
    default_code = 'error'

    def __init__(self, message=None, code=None, **data):
        if message is None:
            message = self.default_message

        if code is None:
            code = self.default_code

        self.code = code
        self.error_data = data

        super().__init__(message)


class NotAuthenticated(GraphQLError):
    default_message = _('Incorrect authentication credentials')
    default_code = 'authenticationFailed'


class PermissionDenied(GraphQLError):
    default_message = _('You do not have permission to perform this action')
    default_code = 'permissionDenied'


class ValidationError(GraphQLError):
    default_message = _('Invalid input')
    default_code = 'invalid'


class NotFound(GraphQLError):
    default_message = _('GraphQL object not found')
    default_code = 'notFound'



from django.core.exceptions import ValidationError
from django.http import Http404
from django.shortcuts import get_object_or_404 as _get_object_or_404

from . import exceptions


def get_object_or_not_found(queryset, *args, **kwargs):
    try:
        return _get_object_or_404(queryset, *args, **kwargs)
    except (TypeError, ValueError, ValidationError, Http404):
        raise exceptions.NotFound(**kwargs)



CHOICES_LOOKUPS = ['exact', 'in']

TEXT_LOOKUPS = CHOICES_LOOKUPS + [
    'iexact', 'contains', 'icontains', 'startswith',
    'istartswith', 'endswith', 'iendswith', 'regex', 'iregex',
]

RANGE_LOOKUPS = ['exact', 'gt', 'gte', 'lt', 'lte']

DATETIME_LOOKUPS = RANGE_LOOKUPS + \
    ['date__' + lookup for lookup in RANGE_LOOKUPS] + \
    ['year', 'month', 'week_day', 'day', 'hour']
